/*
 * input.c
 * 
 * Modified to work as client in socket based protocol
 */
#include "copyright.h"

#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#ifdef hpux
#include <time.h>
#else	/* hpux */
#include <sys/time.h>
#endif				/* hpux */
#include <signal.h>
#include <errno.h>
#include "Wlib.h"
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"

#ifdef FEATURE
static struct timeval curtp;
static struct timeval lasttp;
#endif

#ifdef XSIG
int	Xpending=0;
#endif

initkeymap()
{
   unsigned char  *str;

   if ((str = (unsigned char *) getdefault("keymap")) != NULL) {
      while (*str != '\0') {
	 if (*str >= 32 && *str < 176) {
	    mystats->st_keymap[*str - 32] = *(str + 1);
	 }
	 str += 2;
      }
   }
#ifdef TSH
   /* note: not stored on server */
   if ((str = (unsigned char *) getdefault("buttonmap")) != NULL) {
      while (*str != '\0') {
	 switch (*str++) {
	 case '1':
	    buttonmap[1] = *str;
	    break;
	 case '2':
	    buttonmap[2] = *str;
	    break;
	 case '3':
	    buttonmap[3] = *str;
	    break;
	 default:
	    fprintf(stderr, "%c ignored in buttonmap\n", *(str - 1));
	    break;
	 }
	 str++;
      }
   }
#endif				/* TSH */

}

initinput()
{
   /* Nothing doing... */
}

input()
{
   fd_set          readfds;
#ifndef XSIG
   register
   int		   xsock = W_Socket();	/* put it in a register */
#endif
   register
   int		   doflush=0;

#ifdef XSIG
   xsig_on();
#endif

   while (1) {

#ifdef nodefXSIG
      if(Xpending){
	 Xpending = 0;
	 if(W_ReadEvents()){	/* force read into event queue */
	    if(process_event())
	       W_Flush();
	 }
      }
#endif
      FD_ZERO(&readfds);
#ifndef XSIG
      FD_SET(xsock, &readfds);
#endif
      FD_SET(sock, &readfds);
      if (udpSock >= 0)
	 FD_SET(udpSock, &readfds);

#ifdef XSIG
      if(select(max_fd, &readfds, 0, 0, 0) < 0){
	 continue;
      }
#else
      select(max_fd, &readfds, 0, 0, 0);
#endif

#ifndef XSIG
      if (FD_ISSET(xsock, &readfds)){
	 process_event();
	 /* NOTE: we're no longer calling XPending(), need this */
	 doflush = 1;
      }
#endif

      if (FD_ISSET(sock, &readfds) ||
	  (udpSock >= 0 && FD_ISSET(udpSock, &readfds))) {
	 intrupt(&readfds);
	 /* NOTE: we're no longer calling XPending(), need this */
	 doflush = 1;
	 if (isServerDead()) {
	    printf("Whoops!  We've been ghostbusted!\n");
	    printf("Pray for a miracle!\n");
	    /* UDP fail-safe */
	    commMode = commModeReq = COMM_TCP;
	    commSwitchTimeout = 0;
	    if (udpSock >= 0)
	       closeUdpConn();
	    if (udpWin) {
	       udprefresh(UDP_CURRENT);
	       udprefresh(UDP_STATUS);
	    }
	    connectToServer(nextSocket);
	    printf("Yea!  We've been resurrected!\n");
	 }
      }

      /* NOTE: we're no longer calling XPending(), need this */
      if(doflush){
	 W_Flush();
	 doflush=0;
      }
   }
}

#ifdef XSIG

sigio_handler()
{
   if(critical){
      Xpending = 1;
      return;
   }
   Xpending = 0;

   if(!W_ReadEvents())	/* force read into event queue */
      return;

   if(process_event())
      W_Flush();
}

#endif

process_event()
{
   W_Event         data;
   int		loop =0;

#ifdef XSIG
   while(W_EventsQueued())
#else
   do 
#endif
   {
      loop++;
      if(!W_SpNextEvent(&data)) continue;	/* continues at loop bottom */

      switch ((int) data.type) {
      case W_EV_KEY:
	 if (data.Window == optionWin){
	    optionaction(&data);
	 }
#ifdef NOWARP
	 else if (messageon || data.Window == messagew){
	    smessage(data.key);
	 }
#else
	 else if (data.Window == messagew){
	    smessage(data.key);
	 }
#endif
	 else{
	    keyaction(&data);
	 }
	 break;
   case W_EV_BUTTON:
	 if (data.Window == war){
	    waraction(&data);
	 }
	 else if (data.Window == optionWin){
	    optionaction(&data);
	 }
#ifdef ATM
	 else if (data.Window == udpWin){
	    udpaction(&data);	/* UDP */
#endif				/* ATM */
	 }
	 else
	    buttonaction(&data);
	 break;
      case W_EV_EXPOSE:
	 /* most common first */
	 if (data.Window == mapw){
	    setup_redraw_map(data.x, data.y, 
			     data.x+data.width, data.y+data.height);
#ifdef nodef
	    if(!redrawall)
	       redrawall = 2;	/* no need to clear */
#endif
	 }
	 else if(data.Window == w)	/* nothing special to do here */
	    break;
	 else if (data.Window == statwin)
	    redrawStats();
#ifdef NETSTAT
	 else if (data.Window == lMeter)
	    redrawLMeter();
#endif
#ifdef PING
	 else if (data.Window == pStats)
	    redrawPStats();
#endif
	 else if (data.Window == tstatw)
	    redrawTstats();
	 else if (data.Window == iconWin)
	    drawIcon();
	 else if (data.Window == helpWin)
	    fillhelp();
#ifdef XTREKRC_HELP
	 else if (defWin && (data.Window == defWin))
	    showdef();
#endif
	 else if (data.Window == playerw)
	    playerlist();
	 else if (data.Window == planetw)
	    planetlist();
	 else if (data.Window == rankw)
	    ranklist();
	 else if (data.Window == warnw)
	    W_ClearWindow(warnw);
	 else if (data.Window == messagew)
	    W_ClearWindow(messagew);
#ifdef FEATURE
	 else if (macroWin && data.Window == macroWin)
	    fillmacro();
#endif
	 break;
      default:
	 break;
      }
   } 
#ifndef XSIG
   while(W_EventsQueued());
#endif
   return loop;
}

#ifdef XTRA
xinput()

{
static
   struct timeval  timeout = { 0, 0 };
   fd_set          readfds;
   int             xsock = W_Socket();

   FD_ZERO(&readfds);
   FD_SET(xsock, &readfds);
   if(select(max_fd, &readfds, 0, 0, &timeout) == 1){
      process_event();
   }
}
#endif

keyaction(data)
    W_Event        *data;
{
   unsigned char   course;
   struct obtype  *gettarget(), *target;
   unsigned char   key = data->key;

   if (data->Window != mapw && data->Window != w && (!infow || data->Window != infow)
#ifdef ATM
       && data->Window != scanw
#endif				/* ATM */
      )
      return;
#ifdef nodef
   if (infow && data->Window == infow) {
      int             x, y;
      if (findMouseInWin(&x, &y, w)) {	/* local window */
	 data->Window = w;
	 data->x = x;
	 data->y = y;
      } else if (findMouseInWin(&x, &y, mapw)) {	/* galactic window */
	 data->Window = mapw;
	 data->x = x;
	 data->y = y;
      }
   }
#endif

#ifdef FEATURE
   if(key == macrokey) key = 'X';
#endif

   if (!index("sbogadc", key) || !(localflags & PFREFIT)) {
      if (key >= 32 && key < 176) {
	 key = mystats->st_keymap[key - 32];
      }
   }

#ifdef FEATURE
   if (MacroMode) {
      doMacro(data->key, data);
      return 0;
   }
#endif

   switch(key){
#ifdef FEATURE
   case 'X':
      gettimeofday(&curtp, NULL);
      if (0 /* curtp.tv_sec < lasttp.tv_sec + 6 */) {
	 warning("Chill out with the messages!!");
      } else {
	 warning("In Macro Mode");
	 if(F_UseNewMacro)
	    MacroNum = -1;
	 MacroMode = 1;
      }
      lasttp = curtp;
      break;
#endif

   case '0':
   case '1':
   case '2':
   case '3':
   case '4':
   case '5':
   case '6':
   case '7':
   case '8':
   case '9':
      set_speed(key - '0');
      localflags &= ~(PFREFIT);
      break;
   case 'e':			/* Turn off docking permission, eject docked
				 * vessels. */
      if (me->p_flags & PFDOCKOK)
	 sendDockingReq(0);
      else
	 sendDockingReq(1);
      break;
   case 'r':
      localflags |= PFREFIT;
#ifdef GALAXY
      warning("s=scout, d=destroyer, c=cruiser, b=battleship, a=assault, g=galaxy, o=starbase/outpost");
#else
      warning("s=scout, d=destroyer, c=cruiser, b=battleship, a=assault, o=starbase/outpost");
#endif				/* GALAXY */
      break;
   case ')':
      set_speed(10);
      localflags &= ~(PFREFIT);
      break;
   case '!':
      set_speed(11);
      localflags &= ~(PFREFIT);
      break;
   case '@':
      set_speed(12);
      localflags &= ~(PFREFIT);
      break;
   case '#':
      set_speed(me->p_ship.s_maxspeed / 2);
      localflags &= ~(PFREFIT);
      break;
   case '%':
      set_speed(99);		/* Max speed... */
      localflags &= ~(PFREFIT);
      break;
#ifdef EM
   case '>':
      set_speed(me->p_speed + 1);
      break;
   case '<':
      set_speed(me->p_speed - 1);
      break;
   case '{':
      cloak_on();
      break;
   case '}':
      cloak_off();
      break;
#endif				/* EM */
   case 'a':
      if (localflags & PFREFIT) {
	 do_refit(ASSAULT);
      }
#ifdef ATM
      else {
	 if (!W_IsMapped(scanwin)) {
	    scan(data->Window, data->x, data->y);
	 } else {
	    if (scanmapped)
	       W_UnmapWindow(scanwin);
	    scanmapped = 0;
	 }
      }
#endif				/* ATM */
      break;
#ifdef GALAXY
   case 'g':
      if (localflags & PFREFIT) {
	 do_refit(SGALAXY);
      }
      break;
#endif				/* GALAXY */
   case 'm':			/* new from galaxy -- works here too */
#ifdef NOWARP
      message_on();
#else
      W_WarpPointer(messagew);
#endif
      break;

   case 'k':			/* k = set course */
      course = getcourse(data->Window, data->x, data->y);
      set_course(course);
      me->p_flags &= ~(PFPLOCK | PFPLLOCK);
      localflags &= ~(PFREFIT);
      break;
   case 'p':			/* p = fire phasers */
      course = getcourse(data->Window, data->x, data->y);
      sendPhaserReq(course);
      break;
   case 't':			/* t = launch torps */
      course = getcourse(data->Window, data->x, data->y);
      sendTorpReq(course);
      break;
   case 'f':
      /* f = launch plasma torpedos */
      course = getcourse(data->Window, data->x, data->y);
      sendPlasmaReq(course);
      break;
   case 'd':			/* d = detonate other torps */
      if (localflags & PFREFIT)
	 do_refit(DESTROYER);
      else
	 sendDetonateReq();
      break;
   case 'D':			/* D = detonate my torps */
      detmine();
      break;
   case '[':
      shield_down();
      break;
   case ']':
      shield_up();
      break;
   case 'u':			/* u = toggle shields */
      shield_tog();
      break;
   case 's':			/* For Serge */
      if (localflags & PFREFIT)
	 do_refit(SCOUT);
      else
	 shield_tog();
      break;
   case 'b':			/* b = bomb planet */
      if (localflags & PFREFIT)
	 do_refit(BATTLESHIP);
      else
	 bomb_planet();
      break;
   case 'z':			/* z = beam up */
      beam_up();
      break;
   case 'x':			/* x = beam down */
      beam_down();
      break;
   case 'R':			/* R = Go into repair mode */
      sendRepairReq(1);
      break;
   case 'y':
   case 'T':
      if (me->p_flags & (PFTRACT | PFPRESS)) {
	 sendTractorReq(0, me->p_no);
	 break;
      }
      target = gettarget(data->Window, data->x, data->y, TARG_PLAYER);
#ifdef TSH
      me->p_tractor = target->o_num;
#endif				/* TSH */
      if (key == 'T') {
	 sendTractorReq(1, target->o_num);
      } else {
	 sendRepressReq(1, target->o_num);
      }
      break;
   case 'o':			/* o = dock at nearby starbase or orbit
				 * nearest planet */
      if (localflags & PFREFIT) {
	 do_refit(STARBASE);
      } else {
	 sendOrbitReq(1);
      }
      break;
   case 'O':			/* O = options Window */
      if (optionWin != NULL && W_IsMapped(optionWin))
	 optiondone();
      else
	 optionwindow();
      break;
   case 'Q':
      sendQuitReq();
#ifdef XSIG
      if(sock < 0){
	 printf("Giving up on ghostbust recovery.\n");
	 exit(0);
      }
#endif
      break;
   case 'V':
      showlocal++;
      if (showlocal == 3)
	 showlocal = 0;
      if (optionWin)
	 optionredrawoption(&showlocal);

      break;
   case 'B':
      showgalactic++;
      if (showgalactic == 3)
	 showgalactic = 0;
      if(!redrawall)
	 redrawall = 2;
      if (optionWin)
	  optionredrawoption(&showlocal);
      break;
   case '?':			/* ? = Redisplay all message windows */
      if (!W_IsMapped(reviewWin)) {
	 if (W_IsMapped(messwa)) {
	    W_UnmapWindow(messwa);
	    W_UnmapWindow(messwt);
	    W_UnmapWindow(messwi);
	    W_UnmapWindow(messwk);
	 } else {
	    W_MapWindow(reviewWin);
	 }
      } else {
	 W_UnmapWindow(reviewWin);
	 W_MapWindow(messwa);
	 W_MapWindow(messwt);
	 W_MapWindow(messwi);
	 W_MapWindow(messwk);
      }
      if (optionWin) {
	 optionredrawtarget(reviewWin);
	 optionredrawtarget(messwa);
	 optionredrawtarget(messwt);
	 optionredrawtarget(messwi);
	 optionredrawtarget(messwk);
      }
      break;
   case 'c':			/* c = cloak */
      if (localflags & PFREFIT)
	 do_refit(CRUISER);
      else
	 cloak();
      break;
   case 'C':			/* C = coups */
      sendCoupReq();
      break;
   case 'l':			/* l = lock onto */
      target = gettarget(data->Window, data->x, data->y,
			 TARG_PLAYER | TARG_PLANET);
      if (target->o_type == PLAYERTYPE) {
	 sendPlaylockReq(target->o_num);
#ifdef TSH
	 me->p_playerl = target->o_num;
	 redrawPlayer[me->p_playerl] = 1;
#endif				/* TSH */

      } else {			/* It's a planet */
	 sendPlanlockReq(target->o_num);
#ifdef TSH
	 me->p_planet = target->o_num;
         planets[me->p_planet].pl_flags |= PLREDRAW;
#endif				/* TSH */
      }
      break;
   case '*':
      sendPractrReq();
      break;
      /* Start of display functions */
   case ' ':			/* ' ' = clear special windows */
      W_UnmapWindow(planetw);
      W_UnmapWindow(rankw);
      if (infomapped)
	 destroyInfo();
      W_UnmapWindow(helpWin);
#ifdef XTREKRC_HELP
      if(defWin)
	 W_UnmapWindow(defWin);
#endif
      W_UnmapWindow(war);
      if (optionWin)
	 optiondone();
#ifdef ATM
      if (scanmapped) {
	 W_UnmapWindow(scanwin);
	 scanmapped = 0;
      }
      if (udpWin)
	 udpdone();
#endif				/* ATM */
      break;
   case 'E':			/* E = send emergency call */
      emergency();
      break;
#ifdef EM
   case 'F':			/* F = send armies carried report */
      army_report();
      break;
#endif				/* EM */
   case 'L':			/* L = Player list */
      if (W_IsMapped(playerw)) {
	 W_UnmapWindow(playerw);
      } else {
	 W_MapWindow(playerw);
      }
      break;
   case 'P':			/* P = Planet list */
      if (W_IsMapped(planetw)) {
	 W_UnmapWindow(planetw);
      } else {
	 W_MapWindow(planetw);
      }
      break;
   case 'U':			/* U = Rank list */
      if (W_IsMapped(rankw)) {
	 W_UnmapWindow(rankw);
      } else {
	 W_MapWindow(rankw);
      }
      break;
   case 'S':			/* S = toggle stat mode */
      if (W_IsMapped(statwin)) {
	 W_UnmapWindow(statwin);
      } else {
	 W_MapWindow(statwin);
      }
      break;
#ifdef nodef
   case 'M':			/* M = Toggle Message Logging */
      break;
#endif
   case 'N':			/* N = Toggle Name mode */
      namemode = !namemode;
      if (optionWin)
	 optionredrawoption(&namemode);
      break;
   case 'i':			/* i = get information */
   case 'I':			/* I = get extended information */
      if (!infomapped)
	 inform(data->Window, data->x, data->y, key);
      else
	 destroyInfo();
      break;
   case 'h':			/* h = Map help window */
      if (W_IsMapped(helpWin)) {
	 W_UnmapWindow(helpWin);
      } else {
#ifdef EM
	 fillhelp();
#endif				/* EM */
	 W_MapWindow(helpWin);
      }
      if (optionWin)
	 optionredrawtarget(helpWin);
      break;
   case 'w':			/* w = map war stuff */
      if (W_IsMapped(war))
	 W_UnmapWindow(war);
      else
	 warwindow();
      break;
#ifdef ATM
   case '+':			/* UDP: pop up UDP control window */
      if (udpWin != NULL && W_IsMapped(udpWin))
	 udpdone();
      else {
	 char            buf[80];
	 udpwindow();
	 sprintf(buf, "UDP client version %.1f",
		 (float) UDPVERSION / 10.0);
	 warning(buf);
      }
      if (optionWin)
       optionredrawtarget(udpWin);
      break;
   case '=':			/* UDP: request for full update */
      sendUdpReq(COMM_UPDATE);
      break;
#endif				/* ATM */
#ifdef SHORT_PACKETS
   case '-':
      sendShortReq(SPK_SALL);
      break;
#endif
#ifdef NETSTAT
   case '/':
      netstat = !netstat;
      if(netstat && !W_IsMapped(lMeter)){
	 ns_init(1);
	 W_MapWindow(lMeter);
      }
      else if(!netstat && W_IsMapped(lMeter)){
	 W_UnmapWindow(lMeter);
      }
      break;
#endif				/* NETSTAT */

#ifdef NEW_PL
   case '\\':
      new_playerlist = !new_playerlist;
      W_ClearWindow(playerw);
      playerlist();
      break;
#endif

#ifdef MOO
   case '|':
      sortPlayers = (!sortPlayers);
      if (sortPlayers)
	 warning("Using sorted player list, Captain!");
      else
	 warning("Using regular player list, Captain!");
      W_ClearWindow(playerw);
      playerlist();
      break;
#ifdef LOGMESG
   case 'M':
      logMess = !logMess;
      if (logMess)
	 warning("Message logging is ON, Captain!");
      else
	 warning("Message logging is OFF, Captain!");
      break;
#endif
#ifdef nodef			/* my attempt to control the recorder during
				 * the game <isae> */
   case '`':
      recordGame = !recordGame;
      if (recordGame)
	 warning("Recorder is ON, Captain!");
      else
	 warning("Recorder is OFF, Captain!");
      break;
#endif				/* nodef */
   case ':':			/* Re-read the values from rc file -- only
				 * nifty values are affected */
      warning("Re-reading nifty configurations ...");
      readDefaults();
      break;
#ifdef MOOBITMAPS
   case ';':
      myPlanetBitmap = !myPlanetBitmap;
      if (myPlanetBitmap)
	 warning("Using NEW planet bitmaps ...");
      else
	 warning("Using OLD planet bitmaps ...");
      break;
#endif				/* MOOBITMAPS */

#ifdef PING
   case ',':
      if (W_IsMapped(pStats)) {
	 W_UnmapWindow(pStats);
      } else {
	 W_MapWindow(pStats);
	 redrawPStats();
      }
      break;
#endif

#endif				/* MOO */
   default:
      W_Beep();
      break;
   }
}

buttonaction(data)
    W_Event        *data;
{
   unsigned char   course;
   struct obtype  *gettarget();

#ifdef NOWARP
   if (messageon)
      message_off();		/* ATM */
#endif

#ifdef SHORT_PACKETS
   if (data->Window == reviewWin) {
      if (recv_mesg) {
	 sendShortReq(SPK_MOFF);
      } else {
	 sendShortReq(SPK_MON);
      }
      return;
   }
#endif

   if (data->Window != w && data->Window != mapw
#ifdef TSH
       && (!infow || data->Window != infow)
#endif				/* TSH */
#ifdef ATM
       && data->Window != scanwin
#endif				/* ATM */
      )
      return;

#ifdef TSH
   if (data->key > 0 && data->key <= 3) {
      if (buttonmap[data->key] != '\0') {
	 data->key = buttonmap[data->key];
	 keyaction(data);
	 return;
      }
   }
#endif				/* TSH */

#ifdef TSH
   if (infow && data->Window == infow) {
      int             x, y;
      if (findMouseInWin(&x, &y, w)) {	/* local window */
	 data->Window = w;
	 data->x = x;
	 data->y = y;
      } else if (findMouseInWin(&x, &y, mapw)) {	/* galactic window */
	 data->Window = mapw;
	 data->x = x;
	 data->y = y;
      }
   }
#endif

   if (data->key == W_RBUTTON) {
      course = getcourse(data->Window, data->x, data->y);
      set_course(course);
   } else if (data->key == W_LBUTTON) {
      course = getcourse(data->Window, data->x, data->y);
      sendTorpReq(course);
   } else if (data->key == W_MBUTTON) {
      course = getcourse(data->Window, data->x, data->y);
      sendPhaserReq(course);
   }
}

/*
 * changed from unsigned char to irint() for precise rounding (from Leonard
 * Dickens)
 */
getcourse(ww, x, y)
    W_Window        ww;
    int             x, y;
{
   extern double	rint();
   if (ww == mapw) {
      int             me_x, me_y;

      me_x = me->p_x * WINSIDE / GWIDTH;
      me_y = me->p_y * WINSIDE / GWIDTH;
      return (int)rint(atan2((double) (x - me_x),
			 (double) (me_y - y)) / 3.14159 * 128.);
   } else
      return (int)rint(atan2((double) (x - WINSIDE / 2),
			 (double) (WINSIDE / 2 - y))
		   / 3.14159 * 128.);
}
#ifdef ATM
scan(w, x, y)
{
}
#endif				/* ATM */
